#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
tools.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import cosmosc2


def list_settings():
    return cosmosc2.COSMOS.json_rpc_request("list_settings", [])


def get_all_settings():
    return cosmosc2.COSMOS.json_rpc_request("get_all_settings", [])


def get_setting(name):
    return cosmosc2.COSMOS.json_rpc_request("get_setting", [name])


def get_settings(settings_list):
    return cosmosc2.COSMOS.json_rpc_request("get_settings", settings_list)


def save_setting(name, data):
    return cosmosc2.COSMOS.json_rpc_request("save_setting", [name, data])


def list_configs(tool):
    return cosmosc2.COSMOS.json_rpc_request("list_configs", [tool])


def load_config(tool, name):
    return cosmosc2.COSMOS.json_rpc_request("load_config", [tool, name])


def save_config(tool, name, data):
    return cosmosc2.COSMOS.json_rpc_request("save_config", [tool, name, data])


def delete_config(tool, name):
    return cosmosc2.COSMOS.json_rpc_request("delete_config", [tool, name])


def get_screen_list(config_filename=None, force_refresh=False):
    """Get the organized list of available telemetry screens"""
    return cosmosc2.COSMOS.json_rpc_request(
        "get_screen_list", config_filename, force_refresh
    )


def get_screen_definition(screen_full_name, config_filename=None, force_refresh=False):
    """Get a specific screen definition"""
    return cosmosc2.COSMOS.json_rpc_request(
        "get_screen_definition", screen_full_name, config_filename, force_refresh
    )



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
cosmos_v5_stream_example.py
"""
import argparse
import hashlib
import json
import os
import logging

# See cosmosc2/docs/environment.md for environment documentation

os.environ["COSMOS_VERSION"] = "1.1.1"
os.environ["COSMOS_API_PASSWORD"] = "www"
os.environ["COSMOS_LOG_LEVEL"] = "INFO"
os.environ["COSMOS_WS_SCHEMA"] = "ws"
os.environ["COSMOS_API_HOSTNAME"] = "127.0.0.1"
os.environ["COSMOS_API_PORT"] = "2900"

# os.environ["COSMOS_DATA"] = "\\git\\tmp\\"

from cosmosc2.stream_api.data_extractor_client import DataExtractorClient


def hash_args(args):
    return hashlib.sha1(
        " ".join(
            [" ".join([item for item in args.items]), args.start, args.end, args.mode]
        ).encode("utf-8")
    ).hexdigest()[:16]


def output_data_to_file(data, filename):
    with open(filename, "w") as f:
        f.write("ITEM,VALUE,TIME\n")
        for d in data:
            f.write(f"{d['item']},{d['value']},{d['time']}\n")
    print(filename)


def output_metadata(args, data_path):
    metadata_file = os.path.join(data_path, "metadata.json")
    metadata = {
        "items": args.items,
        "start": args.start,
        "end": args.end,
        "mode": args.mode,
    }
    with open(metadata_file, "w") as f:
        f.write(json.dumps(metadata, indent=4))


def output(args, data, data_path, filename):
    if not data:
        return
    try:
        os.makedirs(data_path)
        output_metadata(args, data_path)
        output_data_to_file(data, filename)
    except OSError as e:
        logging.exception(f"failed to output data, {e}")


# item example: INST.ADCS.POSX
def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "mode", type=str, choices=["CONVERTED", "DECOM", "WITH_UNITS", "RAW"], help="item mode"
    )
    parser.add_argument(
        "start", type=str, help="start time in format: '2022/01/24 11:04:19'"
    )
    parser.add_argument(
        "end", type=str, help="end time in format: '2022/01/25 11:04:19'"
    )
    parser.add_argument(
        "items", type=str, nargs="+", help="item in format: INST.ADCS.POSX"
    )
    args = parser.parse_args()

    dir_name = hash_args(args)
    home = os.path.expanduser("~")
    data_home = os.getenv("COSMOS_DATA", home)
    data_path = os.path.join(data_home, "cosmosc2", dir_name)
    filename = os.path.join(data_path, "data.csv")

    if os.path.isdir(dir_name) and os.path.exists(filename):
        print(filename)
        return

    try:
        api = DataExtractorClient(
            items=args.items,
            start_time=args.start,
            end_time=args.end,
            mode=args.mode,
        )
        data = api.get()
        output(args, data, data_path, filename)
    except ValueError as e:
        logging.error(e)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass
    except Exception as err:
        logging.exception(err)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
json_rpc/request.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import json

from cosmosc2.environment import JSON_RPC_VERSION
from cosmosc2.exceptions import CosmosRequestError
from cosmosc2.json_rpc.base import CosmosJsonRpc


class CosmosJsonRpcRequest(CosmosJsonRpc):
    """Represents a JSON Remote Procedure Call Request"""

    DANGEROUS_METHODS = ["__send__", "send", "instance_eval", "instance_exec"]

    def __init__(self, id_: int, method_name: str, scope: str, *args, **kwargs):
        """Constructor

        Arguments:
        id_ -- The identifier which will be matched to the response
        method_name -- The name of the method to call
        scope -- The scope
        args -- Array of strings which represent the parameters to send to the method
        kwargs -- Dict of Key, Value parameters to send to the method
        """
        super().__init__()
        self["method"] = str(method_name)
        self["params"] = args
        self["keyword_params"] = {**{"scope": scope}, **kwargs}
        self["id"] = int(id_)

    @property
    def method(self):
        """Returns the method to call"""
        return self["method"]

    @property
    def params(self):
        """Returns the array of strings which represent the parameters to send to the method"""
        return self.get("params")

    @property
    def keyword_params(self):
        """Returns a dictionary of strings which represent the keyword parameters to send to the method"""
        return self.get("keyword_params")

    @property
    def id(self):
        """Returns the request identifier"""
        return self["id"]

    def to_hash(self):
        """Returns the request in a string"""
        return _convert_bytearray_to_string_raw(self)

    @classmethod
    def from_json(cls, request_data):
        """Creates and returns a JsonRpcRequest object from a JSON encoded String.

        The version must be 2.0 and the JSON must include the method and id members.

        Parameters:
        request_data -- JSON encoded string representing the request
        """
        msg = "invaid json-rpc {} request".format(JSON_RPC_VERSION)
        try:
            hash_ = json.loads(request_data)
            # Verify the jsonrpc version is correct and there is a method and id
            if hash_["jsonrpc"] != JSON_RPC_VERSION:
                raise ValueError("message jsonrpc version: {}".format(hash_["jsonrpc"]))
            return cls.from_hash(hash_)
        except (ValueError, KeyError) as e:
            raise CosmosRequestError(msg, request_data) from e
        except Exception as e:
            raise RuntimeError(msg) from e

    @classmethod
    def from_hash(cls, hash_):
        """Creates a JsonRpcRequest object from a Hash

        Parameters:
        hash_ -- Hash containing the following keys: method, params, id, and keyword_params
        """
        return cls(
            hash_["id"],
            hash_["method"],
            hash_.get("keyword_params", {}).get("scope"),
            *hash_.get("params", []),
        )


def _convert_bytearray_to_string_raw(object_):
    if isinstance(object_, (bytes, bytearray)):
        return object_.decode("latin-1")
    if isinstance(object_, dict):
        for key, value in object_.items():
            object_[key] = _convert_bytearray_to_string_raw(value)
        return object_
    if isinstance(object_, (tuple, list)):
        object_ = list(object_)
        index = 0
        for value in object_:
            object_[index] = _convert_bytearray_to_string_raw(value)
            index += 1
        return object_
    return object_



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
telemetry.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import cosmosc2


def tlm(*args):

    """Poll for the converted value of a telemetry item
    Usage:
      tlm(target_name, packet_name, item_name)
    or
      tlm('target_name packet_name item_name')
    """
    return cosmosc2.COSMOS.json_rpc_request("tlm", *args)


def tlm_raw(*args):

    """Poll for the raw value of a telemetry item
    Usage:
      tlm_raw(target_name, packet_name, item_name)
    or
      tlm_raw('target_name packet_name item_name')
    """
    return cosmosc2.COSMOS.json_rpc_request("tlm_raw", *args)


def tlm_formatted(*args):

    """Poll for the formatted value of a telemetry item
    Usage:
      tlm_formatted(target_name, packet_name, item_name)
    or
      tlm_formatted('target_name packet_name item_name')
    """
    return cosmosc2.COSMOS.json_rpc_request("tlm_formatted", *args)


def tlm_with_units(*args):

    """Poll for the formatted with units value of a telemetry item
    Usage:
      tlm_with_units(target_name, packet_name, item_name)
    or
      tlm_with_units('target_name packet_name item_name')
    """
    return cosmosc2.COSMOS.json_rpc_request("tlm_with_units", *args)


def tlm_variable(*args):

    return cosmosc2.COSMOS.json_rpc_request("tlm_variable", *args)


def set_tlm(*args):

    """Set a telemetry point to a given value. Note this will be over written in
    a live system by incoming new telemetry.
    Usage:
      set_tlm(target_name, packet_name, item_name, value)
    or
      set_tlm("target_name packet_name item_name = value")
    """
    return cosmosc2.COSMOS.json_rpc_request("set_tlm", *args)


def set_tlm_raw(*args):

    """Set the raw value of a telemetry point to a given value. Note this will
    be over written in a live system by incoming new telemetry.
    Usage:
      set_tlm_raw(target_name, packet_name, item_name, value)
    or
      set_tlm_raw("target_name packet_name item_name = value")
    """
    return cosmosc2.COSMOS.json_rpc_request("set_tlm_raw", *args)


def inject_tlm(
    target_name,
    packet_name,
    item_hash=None,
    value_type="CONVERTED",
    send_routers=True,
    send_packet_log_writers=True,
    create_new_logs=False,
):
    """Injects a packet into the system as if it was received from an interface"""
    return cosmosc2.COSMOS.json_rpc_request(
        "inject_tlm",
        target_name,
        packet_name,
        item_hash,
        value_type,
        send_routers,
        send_packet_log_writers,
        create_new_logs,
    )


def override_tlm(*args):
    """Permanently set the converted value of a telemetry point to a given value
    Usage:
      override_tlm(target_name, packet_name, item_name, value)
    or
      override_tlm("target_name packet_name item_name = value")
    """
    return cosmosc2.COSMOS.json_rpc_request("override_tlm", *args)


def override_tlm_raw(*args):
    """Permanently set the raw value of a telemetry point to a given value
    Usage:
      override_tlm_raw(target_name, packet_name, item_name, value)
    or
      override_tlm_raw("target_name packet_name item_name = value")
    """
    return cosmosc2.COSMOS.json_rpc_request("override_tlm_raw", *args)


def normalize_tlm(*args):
    """Clear an override of a telemetry point
    Usage:
      normalize_tlm(target_name, packet_name, item_name)
    or
      normalize_tlm("target_name packet_name item_name")
    """
    return cosmosc2.COSMOS.json_rpc_request("normalize_tlm", *args)


def get_telemetry(target_name, packet_name):
    """
    The get_telemetry function returns a packet hash.
    Usage:
      packet_hash = get_telemetry(target_name, packet_name)
    """
    return cosmosc2.COSMOS.json_rpc_request("get_telemetry", target_name, packet_name)


def get_tlm_packet(target_name, packet_name, value_type="CONVERTED"):
    """Gets all the values from the given packet returned in a two dimensional
    array containing the item_name, value, and limits state.
    Usage:
      values = get_tlm_packet(target_name, packet_name, <:RAW, :CONVERTED, :FORMATTED, :WITH_UNITS>)
    """
    return cosmosc2.COSMOS.json_rpc_request(
        "get_tlm_packet", target_name, packet_name, type=value_type
    )


def get_tlm_values(items, value_type="CONVERTED"):
    """Gets all the values from the given packet returned in an
    array consisting of an Array of item values, an array of item limits state
    given as symbols such as :RED, :YELLOW, :STALE, an array of arrays including
    the limits setting such as red low, yellow low, yellow high, red high and
    optionally green low and high, and the overall limits state of the system.
    Usage:
      values = get_tlm_values([[target_name, packet_name, item_name], ...], <:RAW, :CONVERTED, :FORMATTED, :WITH_UNITS>)
    """
    return cosmosc2.COSMOS.json_rpc_request("get_tlm_values", items)


def get_target(target_name):
    """
    The get_target method returns a target hash containing all the information about the target.
    """
    return cosmosc2.COSMOS.json_rpc_request("get_target", target_name)


def get_target_list():
    """Gets the list of all defined targets."""
    return cosmosc2.COSMOS.json_rpc_request("get_target_list")


def get_tlm_buffer(target_name, packet_name):
    """The get_tlm_buffer method returns the raw packet buffer as a Ruby string.
    Syntax:
      buffer = get_tlm_buffer("<Target Name>", "<Packet Name>")
    """
    return cosmosc2.COSMOS.json_rpc_request("get_tlm_buffer", target_name, packet_name)


def subscribe_packets(packets: list):
    """The subscribe_packets method allows the user to listen for one or more telemetry packets of data to arrive. A unique id is returned which is used to retrieve the data.
    Syntax:
        id = subscribe_packets([['INST', 'HEALTH_STATUS'], ['INST', 'ADCS']])
    """
    return cosmosc2.COSMOS.json_rpc_request("subscribe_packets", packets)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
internal_api.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import cosmosc2


def cosmos_status():
    """Get the cosmos status api.
    Syntax / Example:
        status = cosmos_status()
    """
    resp = cosmosc2.COSMOS.get(
        "/cosmos-api/internal/status", headers={"Accept": "application/json"}
    )
    return resp.json()


def cosmos_health():
    """Get the cosmos health api.
    Syntax / Example:
        health = cosmos_health()
    """
    resp = cosmosc2.COSMOS.get(
        "/cosmos-api/internal/health", headers={"Accept": "application/json"}
    )
    return resp.json()


def cosmos_metrics():
    """Get the cosmos metrics api.
    Syntax / Example:
        metrics = cosmos_metrics()
    """
    resp = cosmosc2.COSMOS.get(
        "/cosmos-api/internal/metrics", headers={"Accept": "plain/txt"}
    )
    return resp.text



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
json_rpc/error.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

from cosmosc2.environment import JSON_RPC_VERSION


class CosmosJsonRpcError(dict):
    """Represents a JSON Remote Procedure Call Error"""

    def __init__(self, code, message, data=None):
        """Constructor

        Parameters:
        code -- The error type that occurred
        message -- A short description of the error
        data -- Additional information about the error
        """
        super().__init__()
        self["code"] = code
        self["message"] = message
        self["data"] = data

    @property
    def code(self):
        """Returns the error type that occurred"""
        return self.get("code")

    @property
    def message(self):
        """Returns a short description of the error"""
        return self.get("message")

    @property
    def data(self):
        """Returns additional information about the error"""
        return self.get("data")

    @classmethod
    def from_hash(cls, hash_):
        """Creates a JsonRpcError object from a Hash

        Parameters:
        hash -- Hash containing the following keys: code, message, and optionally data
        """
        try:
            code = int(hash_["code"])
            return cls(code, hash_["message"], hash_["data"])
        except ValueError as err:
            error = "Invalid JSON-RPC {}".format(JSON_RPC_VERSION)
            raise RuntimeError("{} {}: {}".format(error, type(err), err)) from err



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
environment.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import os

from cosmosc2.__version__ import __title__, __version__

_cosmos_version = "COSMOS_VERSION"

_cosmos_api_user = "COSMOS_API_USER"

_cosmos_api_client = "COSMOS_API_CLIENT"

_cosmos_api_password = "COSMOS_API_PASSWORD"

_default_scope = "COSMOS_SCOPE"

_cosmos_api_schema = "COSMOS_API_SCHEMA"

_default_ws_schema = "COSMOS_WS_SCHEMA"

_cosmos_api_hostname = "COSMOS_API_HOSTNAME"

_cosmos_api_port = "COSMOS_API_PORT"

_json_rpc_version = "COSMOS_JSON_RPC_VERSION"

_log_level = "COSMOS_LOG_LEVEL"

_max_retry_count = "COSMOS_MAX_RETRY_COUNT"

_user_agent = "COSMOS_USER_AGENT"

COSMOS_API_SCHEMA = os.environ.get(_cosmos_api_schema, "http")

COSMOS_WS_SCHEMA = os.environ.get(_default_ws_schema, "ws")

COSMOS_API_HOSTNAME = os.environ.get(_cosmos_api_hostname, "127.0.0.1")

try:
    COSMOS_API_PORT = int(os.environ.get(_cosmos_api_port))
except TypeError:
    COSMOS_API_PORT = 2900

COSMOS_SCOPE = os.environ.get(_default_scope, "DEFAULT")

COSMOS_API_USER = os.environ.get(_cosmos_api_user)

COSMOS_API_CLIENT = os.environ.get(_cosmos_api_client, "api")

COSMOS_API_PASSWORD = os.environ.get(_cosmos_api_password, "SuperSecret")

COSMOS_VERSION = os.environ.get(_cosmos_version)

JSON_RPC_VERSION = os.environ.get(_json_rpc_version, "2.0")

LOG_LEVEL = os.environ.get(_log_level, "INFO")

try:
    MAX_RETRY_COUNT = int(os.environ.get(_max_retry_count))
except TypeError:
    MAX_RETRY_COUNT = 3

_default_user_agent = [
    f"{__title__}:{__version__}:{JSON_RPC_VERSION}:({COSMOS_VERSION})",
]

if COSMOS_API_USER is not None:
    _default_user_agent[0] += f":({COSMOS_API_USER})"

if os.name == "nt":
    _default_user_agent.append(
        f"{os.environ.get('COMPUTERNAME')}:{os.environ.get('USERNAME')}"
    )
else:
    _default_user_agent.append(f"{os.environ.get('HOSTNAME')}:{os.environ.get('USER')}")

USER_AGENT = os.environ.get(_user_agent, " ".join(_default_user_agent))



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
__init__.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import os
import logging
from cosmosc2.environment import LOG_LEVEL
from cosmosc2.connection import CosmosConnection

logging.basicConfig(
    format="%(asctime)s %(levelname)s %(name)s %(message)s",
    level=logging.getLevelName(LOG_LEVEL),
)

COSMOS = CosmosConnection()


def update_scope(scope: str):
    """Update the Cosmos scope selection

    Parameters:
        scope (str): The scope to use on Cosmos v5
    """
    global COSMOS
    COSMOS.scope = str(scope)
    os.environ["COSMOS_SCOPE"] = str(scope)


def initialize_connection(hostname: str = None, port: int = None):
    """Generate the current session with Cosmos

    Parameters:
        hostname (str): The hostname to connect to Cosmos v5
        port (int): The port to connect to Cosmos v5
    """
    global COSMOS

    if COSMOS:
        COSMOS.shutdown()

    if hostname and port:
        COSMOS = CosmosConnection(hostname, port)
    else:
        COSMOS = CosmosConnection()


def shutdown():
    """Shutdown the current session with Cosmos"""
    global COSMOS
    COSMOS.shutdown()


from cosmosc2.api_shared import *
from cosmosc2.cosmos_api import *
from cosmosc2.commands import *
from cosmosc2.extract import *
from cosmosc2.internal_api import *
from cosmosc2.limits import *
from cosmosc2.telemetry import *
from cosmosc2.timeline_api import *
from cosmosc2.tools import *



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
json_rpc/__init__.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

from cosmosc2.json_rpc.error import CosmosJsonRpcError
from cosmosc2.json_rpc.request import CosmosJsonRpcRequest
from cosmosc2.json_rpc.response import (
    CosmosJsonRpcResponse,
    CosmosJsonRpcErrorResponse,
    CosmosJsonRpcSuccessResponse,
    convert_json_class,
)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
keycloak.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import time
import logging
import requests

from cosmosc2.__version__ import __title__
from cosmosc2.environment import *
from cosmosc2.decorators import request_wrapper, retry_wrapper

LOGGER = logging.getLogger(__title__)


def generate_auth():
    """
    Pick Auth class base on environment variables
    """
    if COSMOS_API_USER is None:
        return CosmosAuthorization()
    return CosmosKeycloakAuthorization()


class CosmosAuthorization(requests.auth.AuthBase):
    """Class to hold token for COSMOS

    The CosmosAuthorization can be used to call COSMOS server methods directly:
      auth = CosmosAuthorization()
      requests.get("example.org", auth=auth)
    """

    def get(self):
        return COSMOS_API_PASSWORD

    def __call__(self, r: requests.Request):
        r.headers["Authorization"] = COSMOS_API_PASSWORD
        return r

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __str__(self):
        return f"<{self.__class__.__name__}>"


class CosmosKeycloakAuthorization(CosmosAuthorization):
    """Class to generate Keycloak token for COSMOS Enterprise
    https://developers.redhat.com/blog/2020/01/29/api-login-and-jwt-token-generation-using-keycloak

    The CosmosKeycloakAuthorization can be used to call the COSMOS keycloak server methods directly:
      auth = CosmosKeycloakAuthorization(schema="", hostname="127.0.0.1", port=7777)
      requests.get("example.org", auth=auth)
    """

    def __init__(
        self,
        schema: str = COSMOS_API_SCHEMA,
        hostname: str = COSMOS_API_HOSTNAME,
        port: int = COSMOS_API_PORT,
    ):
        """Constructor

        Parameters:
        schema -- The schema to connect to cosmos with
        hostname -- The name of the machine which has started the JSON service
        port -- The port number of the JSON service
        """
        self.request_url = f"{schema}://{hostname}:{port}"
        self.refresh_token = None
        self.expires_at = None
        self.refresh_expires_at = None
        self.token = None
    
    def _time_logic(self):
        current_time = time.time()
        if self.token is None:
            self._make_token()
        elif self.refresh_expires_at < current_time:
            self._make_token()
        elif self.expires_at < current_time:
            self._refresh_token()

    def get(self):
        self._time_logic()
        return f"Bearer {self.token}"

    def __call__(self, r: requests.Request):
        self._time_logic()
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r

    def _make_token(self):
        """
        {
            "access_token": "",
            "expires_in": 600,
            "refresh_expires_in": 1800,
            "refresh_token": "",
            "token_type": "bearer",
            "id_token": "",
            "not-before-policy": 0,
            "session_state": "",
            "scope": "openid email profile"
        }
        """
        oath = self._make_token_request().json()
        self.token = oath["access_token"]
        self.refresh_token = oath["refresh_token"]
        current_time = time.time()
        self.expires_at = current_time + oath["expires_in"]
        self.refresh_expires_at = current_time + oath["refresh_expires_in"]

    @request_wrapper
    @retry_wrapper
    def _make_token_request(self):
        """Use the python requests libary to request a token from Cosmos Keycloak.

        This is an internal method that uses two decorators. request_wrapper
        captures errors from the request libary. retry_wrapper captures exceptions
        from the requst_wrapper and will retry or bubble up the error.
        return -- request.Response
            https://docs.python-requests.org/en/master/user/quickstart/#json-response-content
        """
        request_kwargs = {
            "url": f"{self.request_url}/auth/realms/COSMOS/protocol/openid-connect/token",
            "data": f"username={COSMOS_API_USER}&password={COSMOS_API_PASSWORD}&client_id={COSMOS_API_CLIENT}&grant_type=password&scope=openid",
            "headers": {
                "User-Agent": USER_AGENT,
                "Content-Type": "application/x-www-form-urlencoded",
            },
        }
        LOGGER.debug("calling with %s", request_kwargs)
        resp = requests.post(**request_kwargs)
        LOGGER.debug(
            "resp: %s total_seconds: %f content: %s",
            resp,
            resp.elapsed.total_seconds(),
            resp.content,
        )
        return resp

    def _refresh_token(self):
        """
        {
            "access_token": "",
            "expires_in": 600,
            "refresh_expires_in": 1800,
            "refresh_token": "",
            "token_type": "bearer",
            "id_token": "",
            "not-before-policy": 0,
            "session_state": "",
            "scope": "openid email profile"
        }
        """
        oath = self._make_refresh_request().json()
        self.token = oath["access_token"]
        self.refresh_token = oath["refresh_token"]
        current_time = time.time()
        self.expires_at = current_time + oath["expires_in"]
        self.refresh_expires_at = current_time + oath["refresh_expires_in"]

    @request_wrapper
    @retry_wrapper
    def _make_refresh_request(self):
        """Use the python requests libary to refresh the token.

        This is an internal method that uses two decorators. request_wrapper
        captures errors from the request libary. retry_wrapper captures exceptions
        from the requst_wrapper and will retry or bubble up the error.
        return -- request.Response
            https://docs.python-requests.org/en/master/user/quickstart/#json-response-content
        """
        request_kwargs = {
            "url": f"{self.request_url}/auth/realms/COSMOS/protocol/openid-connect/token",
            "data": f"client_id={COSMOS_API_CLIENT}&grant_type=refresh_token&refresh_token={self.refresh_token}",
            "headers": {
                "User-Agent": USER_AGENT,
                "Content-Type": "application/x-www-form-urlencoded",
            },
        }
        LOGGER.debug("calling with %s", request_kwargs)
        resp = requests.post(**request_kwargs)
        LOGGER.debug(
            "resp: %s total_seconds: %f content: %s",
            resp,
            resp.elapsed.total_seconds(),
            resp.content,
        )
        return resp



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
commands.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import logging
import sys

import cosmosc2
from cosmosc2.__version__ import __title__
from cosmosc2.exceptions import CosmosResponseError
from cosmosc2.extract import convert_to_value

LOGGER = logging.getLogger(__title__)


# This is in System.commands in Ruby
def build_cmd_output_string(target_name, cmd_name, cmd_params, raw=False):
    if raw:
        output_string = 'cmd_raw("'
    else:
        output_string = 'cmd("'
    output_string += target_name + " " + cmd_name
    if cmd_params is None or len(cmd_params) == 0:
        output_string += '")'
    else:
        params = []
        for key, value in cmd_params.items():
            if isinstance(value, str):
                if isinstance(convert_to_value(value), str):
                    value = repr(value)
                    if len(value) > 256:
                        value = value[0:256] + "...'"
                    value = value.replace('"', "'")
            params.append("{:s} {:s}".format(key, str(value)))
        params = ", ".join(params)
        output_string += " with " + params + '")'
    return output_string


def prompt_for_hazardous(target_name, cmd_name, hazardous_description):
    """ """
    message_list = [
        "Warning: Command {:s} {:s} is Hazardous. ".format(target_name, cmd_name)
    ]
    if hazardous_description:
        message_list.append(" >> {:s}".format(hazardous_description))
    message_list.append("Send? (y/N): ")
    answer = input("\n".join(message_list))
    try:
        return answer.lower()[0] == "y"
    except IndexError:
        return False


def prompt_for_script_abort():
    """ """
    answer = input("Stop running script? (y/N): ")
    try:
        if answer.lower()[0] == "y":
            sys.exit(66)  # execute order 66
    except IndexError:
        return False


def _log_cmd(cmd_, target_name, cmd_name, cmd_params):
    """Log any warnings about disabling checks and log the command itself
    NOTE: This is a helper method and should not be called directly"""
    cmd_str = build_cmd_output_string(target_name, cmd_name, cmd_params, "raw" in cmd_)
    if "no_range" in cmd_ or "no_checks" in cmd_:
        LOGGER.warning(f"{cmd_str} being sent ignoring range checks")
    if "no_hazardous" in cmd_ or "no_checks" in cmd_:
        LOGGER.warning(f"{cmd_str} being sent ignoring hazardous warnings")


def _cmd(cmd_, cmd_no_hazardous, *args):
    """Send the command and log the results
    NOTE: This is a helper method and should not be called directly"""
    try:
        target_name, cmd_name, cmd_params = cosmosc2.COSMOS.json_rpc_request(
            cmd_, *args
        )
        _log_cmd(cmd_, target_name, cmd_name, cmd_params)
    except CosmosResponseError as error:
        resp_error = error.response.error().data()["instance_variables"]
        ok_to_proceed = prompt_for_hazardous(
            resp_error["@target_name"],
            resp_error["@cmd_name"],
            resp_error["@hazardous_description"],
        )
        if ok_to_proceed:
            target_name, cmd_name, cmd_params = cosmosc2.COSMOS.json_rpc_request(
                cmd_no_hazardous, *args
            )
            _log_cmd(cmd_no_hazardous, target_name, cmd_name, cmd_params)
        else:
            prompt_for_script_abort()


def cmd(*args):
    """Send a command to the specified target
    Usage:
      cmd(target_name, cmd_name, cmd_params = {})
    or
      cmd('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd", *args)


def cmd_no_range_check(*args):
    """Send a command to the specified target without range checking parameters
    Usage:
      cmd_no_range_check(target_name, cmd_name, cmd_params = {})
    or
      cmd_no_range_check('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_no_range_check", *args)


def cmd_no_hazardous_check(*args):
    """Send a command to the specified target without hazardous checks
    Usage:
      cmd_no_hazardous_check(target_name, cmd_name, cmd_params = {})
    or
      cmd_no_hazardous_check('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_no_hazardous_check", *args)


def cmd_no_checks(*args):
    """Send a command to the specified target without range checking or hazardous checks
    Usage:
      cmd_no_checks(target_name, cmd_name, cmd_params = {})
    or
      cmd_no_checks('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_no_checks", *args)


def cmd_raw(*args):
    """Send a command to the specified target without running conversions
    Usage:
      cmd_raw(target_name, cmd_name, cmd_params = {})
    or
      cmd_raw('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_raw", *args)


def cmd_raw_no_range_check(*args):
    """Send a command to the specified target without range checking parameters or running conversions
    Usage:
      cmd_raw_no_range_check(target_name, cmd_name, cmd_params = {})
    or
      cmd_raw_no_range_check('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_raw_no_range_check", *args)


def cmd_raw_no_hazardous_check(*args):
    """Send a command to the specified target without hazardous checks or running conversions
    Usage:
      cmd_raw_no_hazardous_check(target_name, cmd_name, cmd_params = {})
    or
      cmd_raw_no_hazardous_check('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_raw_no_hazardous_check", None, *args)


def cmd_raw_no_checks(*args):
    """Send a command to the specified target without range checking or hazardous checks or running conversions
    Usage:
      cmd_raw_no_checks(target_name, cmd_name, cmd_params = {})
    or
      cmd_raw_no_checks('target_name cmd_name with cmd_param1 value1, cmd_param2 value2')
    """
    return _cmd("cmd_raw_no_checks", None, *args)


def send_raw(interface_name, data):
    """Sends raw data through an interface"""
    return cosmosc2.COSMOS.json_rpc_request("send_raw", interface_name, data)


def send_raw_file(interface_name, filename):
    """Sends raw data through an interface from a file"""
    with open(filename, "rb") as file:
        data = file.read()
    return cosmosc2.COSMOS.json_rpc_request("send_raw", interface_name, data)


def get_cmd_list(target_name):
    """Returns all the target commands as an array of arrays listing the command name and description."""
    return cosmosc2.COSMOS.json_rpc_request("get_cmd_list", target_name)


def get_cmd_param_list(target_name, cmd_name):
    """Returns all the parameters for given command as an array of arrays
    containing the parameter name, default value, states, description, units
    full name, units abbreviation, and whether it is required."""
    return cosmosc2.COSMOS.json_rpc_request("get_cmd_param_list", target_name, cmd_name)


def get_cmd_hazardous(target_name, cmd_name, cmd_params=None):
    """Returns whether a command is hazardous (true or false)"""
    if cmd_params is None:
        cmd_params = {}
    return cosmosc2.COSMOS.json_rpc_request(
        "get_cmd_hazardous", target_name, cmd_name, cmd_params
    )


def get_cmd_value(target_name, command_name, parameter_name, value_type="CONVERTED"):
    """Returns a value from the specified command"""
    return cosmosc2.COSMOS.json_rpc_request(
        "get_cmd_value", target_name, command_name, parameter_name, value_type
    )


def get_cmd_time(target_name=None, command_name=None):
    """Returns the time the most recent command was sent"""
    return cosmosc2.COSMOS.json_rpc_request("get_cmd_time", target_name, command_name)


def get_cmd_buffer(target_name, command_name):
    """Returns the buffer from the most recent specified command"""
    return cosmosc2.COSMOS.json_rpc_request("get_cmd_buffer", target_name, command_name)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
api_shared.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt


from cosmosc2.environment import LOG_LEVEL
import time
import logging

from cosmosc2.__version__ import __title__
from cosmosc2 import extract, telemetry
from cosmosc2.exceptions import CosmosCheckError

DEFAULT_TLM_POLLING_RATE = 0.25
LOGGER = logging.getLogger(__title__)


def _upcase(target_name, packet_name, item_name):
    """Creates a string with the parameters upcased"""
    return "{:s} {:s} {:s}".format(
        target_name.upper(), packet_name.upper(), item_name.upper()
    )


def _check(method, *args):
    """Implementaiton of the various check commands. It yields back to the
    caller to allow the return of the value through various telemetry calls.
    This method should not be called directly by application code."""
    target_name, packet_name, item_name, comparison_to_eval = check_process_args(
        args, "check"
    )
    value = method(target_name, packet_name, item_name)
    if comparison_to_eval:
        return check_eval(
            target_name, packet_name, item_name, comparison_to_eval, value
        )
    else:
        LOGGER.info(
            "CHECK: %s == %s", _upcase(target_name, packet_name, item_name), str(value)
        )


def check(*args):
    """Check the converted value of a telmetry item against a condition
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check(target_name, packet_name, item_name, comparison_to_eval)
    or
    check('target_name packet_name item_name > 1')
    """
    return _check(telemetry.tlm, *args)


def check_formatted(*args):
    """Check the formatted value of a telmetry item against a condition
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check(target_name, packet_name, item_name, comparison_to_eval)
    or
    check('target_name packet_name item_name > 1')
    """
    return _check(telemetry.tlm_formatted, *args)


def check_with_units(*args):
    """Check the formatted with units value of a telmetry item against a condition
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check(target_name, packet_name, item_name, comparison_to_eval)
    or
    check('target_name packet_name item_name > 1')
    """
    return _check(telemetry.tlm_with_units, *args)


def check_raw(*args):
    """Check the raw value of a telmetry item against a condition
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check(target_name, packet_name, item_name, comparison_to_eval)
    or
    check('target_name packet_name item_name > 1')
    """
    return _check(telemetry.tlm_raw, *args)


def _check_tolerance(method, *args):
    (
        target_name,
        packet_name,
        item_name,
        expected_value,
        tolerance,
    ) = check_tolerance_process_args(args, "check_tolerance")
    value = method(target_name, packet_name, item_name)
    if isinstance(value, list):
        expected_value, tolerance = array_tolerance_process_args(
            len(value), expected_value, tolerance, "check_tolerance"
        )

        message = ""
        all_checks_ok = True
        for i in range(len(value)):
            range_bottom = expected_value[i] - tolerance[i]
            range_top = expected_value[i] + tolerance[i]
            check_str = "CHECK: {:s}[{:d}]".format(
                _upcase(target_name, packet_name, item_name), i
            )
            range_str = "range {:g} to {:g} with value == {:g}".format(
                range_bottom, range_top, value[i]
            )
            if value[i] >= range_bottom and value[i] <= range_top:
                message += "{:s} was within #{:s}\n".format(check_str, range_str)
            else:
                message += "{:s} failed to be within {:s}\n".format(
                    check_str, range_str
                )
                all_checks_ok = False

        if all_checks_ok:
            LOGGER.info(message)
        else:
            raise CosmosCheckError(message)
    else:
        range_bottom = expected_value - tolerance
        range_top = expected_value + tolerance
        check_str = "CHECK: {:s}".format(_upcase(target_name, packet_name, item_name))
        range_str = "range {:g} to {:g} with value == {:g}".format(
            range_bottom, range_top, value
        )
        if value >= range_bottom and value <= range_top:
            LOGGER.info("{:s} was within {:s}".format(check_str, range_str))
        else:
            message = "{:s} failed to be within {:s}".format(check_str, range_str)
            raise CosmosCheckError(message)


def check_tolerance(*args):
    """Check the converted value of a telmetry item against an expected value with a tolerance
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check_tolerance(target_name, packet_name, item_name, expected_value, tolerance)
    or
    check_tolerance('target_name packet_name item_name', expected_value, tolerance)
    """
    return _check_tolerance(telemetry.tlm, *args)


def check_tolerance_raw(*args):
    """Check the raw value of a telmetry item against an expected value with a tolerance
    Always print the value of the telemetry item to STDOUT
    If the condition check fails, raise an error
    Supports two signatures:
    check_tolerance_raw(target_name, packet_name, item_name, expected_value, tolerance)
    or
    check_tolerance_raw('target_name packet_name item_name', expected_value, tolerance)
    """
    return _check_tolerance(telemetry.tlm_raw, *args)


def check_expression(exp_to_eval, locals=None):
    """Check to see if an expression is true without waiting.  If the expression
    is not true, the script will pause."""
    success = cosmos_script_wait_implementation_expression(
        exp_to_eval, 0, DEFAULT_TLM_POLLING_RATE, locals
    )
    if success:
        LOGGER.info("CHECK: {:s} is TRUE".format(exp_to_eval))
    else:
        message = "CHECK: {:s} is FALSE".format(exp_to_eval)
        raise CosmosCheckError(message)


def wait(*args):
    """Wait on an expression to be true.  On a timeout, the script will continue.
    Supports multiple signatures:
    wait(time)
    wait('target_name packet_name item_name > 1', timeout, polling_rate)
    wait('target_name', 'packet_name', 'item_name', comparison_to_eval, timeout, polling_rate)
    """
    wait_process_args(args, "wait", "CONVERTED")


def wait_raw(*args):
    """Wait on an expression to be true.  On a timeout, the script will continue.
    Supports multiple signatures:
    wait(time)
    wait_raw('target_name packet_name item_name > 1', timeout, polling_rate)
    wait_raw('target_name', 'packet_name', 'item_name', comparison_to_eval, timeout, polling_rate)"""
    wait_process_args(args, "wait_raw", "RAW")


def _wait_tolerance(raw, *args):
    if raw:
        type = "RAW"
    else:
        type = "CONVERTED"
    type_string = "wait_tolerance"
    if raw:
        type_string += "_raw"
    (
        target_name,
        packet_name,
        item_name,
        expected_value,
        tolerance,
        timeout,
        polling_rate,
    ) = wait_tolerance_process_args(args, type_string)
    start_time = time.time()
    value = telemetry.tlm_variable(target_name, packet_name, item_name, type)
    if isinstance(value, list):
        expected_value, tolerance = array_tolerance_process_args(
            len(value), expected_value, tolerance, type_string
        )
        success, value = cosmos_script_wait_implementation_array_tolerance(
            len(value),
            target_name,
            packet_name,
            item_name,
            type,
            expected_value,
            tolerance,
            timeout,
            polling_rate,
        )
        time_float = time.time() - start_time

        message = ""
        for i in range(0, len(value)):
            range_bottom = expected_value[i] - tolerance[i]
            range_top = expected_value[i] + tolerance[i]
            check_str = "WAIT: {:s}[{:d}]".format(
                _upcase(target_name, packet_name, item_name), i
            )
            range_str = "range {:g} to {:g} with value == {:g} after waiting {:g} seconds".format(
                range_bottom, range_top, value[i], time_float
            )
            if value[i] >= range_bottom and value[i] <= range_top:
                message += "{:s} was within #{:s}\n".format(check_str, range_str)
            else:
                message += "{:s} failed to be within {:s}\n".format(
                    check_str, range_str
                )

        if success:
            LOGGER.info(message)
        else:
            LOGGER.warn(message)
    else:
        success, value = cosmos_script_wait_implementation_tolerance(
            target_name,
            packet_name,
            item_name,
            type,
            expected_value,
            tolerance,
            timeout,
            polling_rate,
        )
        time_float = time.time() - start_time
        range_bottom = expected_value - tolerance
        range_top = expected_value + tolerance
        wait_str = "WAIT: {:s}".format(_upcase(target_name, packet_name, item_name))
        range_str = (
            "range {:g} to {:g} with value == {:g} after waiting {:g} seconds".format(
                range_bottom, range_top, value, time_float
            )
        )
        if success:
            LOGGER.info("{:s} was within {:s}".format(wait_str, range_str))
        else:
            LOGGER.warning("{:s} failed to be within {:s}".format(wait_str, range_str))
    return time_float


def wait_tolerance(*args):
    """Wait on an expression to be true.  On a timeout, the script will continue.
    Supports multiple signatures:
    wait_tolerance('target_name packet_name item_name', expected_value, tolerance, timeout, polling_rate)
    wait_tolerance('target_name', 'packet_name', 'item_name', expected_value, tolerance, timeout, polling_rate)
    """
    return _wait_tolerance(False, *args)


def wait_tolerance_raw(*args):
    """Wait on an expression to be true.  On a timeout, the script will continue.
    Supports multiple signatures:
    wait_tolerance_raw('target_name packet_name item_name', expected_value, tolerance, timeout, polling_rate)
    wait_tolerance_raw('target_name', 'packet_name', 'item_name', expected_value, tolerance, timeout, polling_rate)
    """
    return _wait_tolerance(True, *args)


def wait_expression(
    exp_to_eval, timeout, polling_rate=DEFAULT_TLM_POLLING_RATE, locals=None
):
    """Wait on a custom expression to be true"""
    start_time = time.time()
    success = cosmos_script_wait_implementation_expression(
        exp_to_eval, timeout, polling_rate, locals
    )
    time_float = time.time() - start_time
    if success:
        LOGGER.info(
            "WAIT: {:s} is TRUE after waiting {:g} seconds".format(
                exp_to_eval, time_float
            )
        )
    else:
        LOGGER.warning(
            "WAIT: {:s} is FALSE after waiting {:g} seconds".format(
                exp_to_eval, time_float
            )
        )
    return time_float


def _wait_check(raw, *args):
    if raw:
        type = "RAW"
    else:
        type = "CONVERTED"
    (
        target_name,
        packet_name,
        item_name,
        comparison_to_eval,
        timeout,
        polling_rate,
    ) = wait_check_process_args(args, "wait_check")
    start_time = time.time()
    success, value = cosmos_script_wait_implementation(
        target_name,
        packet_name,
        item_name,
        type,
        comparison_to_eval,
        timeout,
        polling_rate,
    )
    time_float = time.time() - start_time
    check_str = "CHECK: {:s} {:s}".format(
        _upcase(target_name, packet_name, item_name), comparison_to_eval
    )
    with_value_str = "with value == {:s} after waiting {:g} seconds".format(
        str(value), time_float
    )
    if success:
        LOGGER.info("{:s} success {:s}".format(check_str, with_value_str))
    else:
        message = "{:s} failed {:s}".format(check_str, with_value_str)
        raise CosmosCheckError(message)
    return time_float


def wait_check(*args):
    """Wait for the converted value of a telmetry item against a condition or for a timeout
    and then check against the condition
    Supports two signatures:
    wait_check(target_name, packet_name, item_name, comparison_to_eval, timeout, polling_rate)
    or
    wait_check('target_name packet_name item_name > 1', timeout, polling_rate)"""
    return _wait_check(False, *args)


def wait_check_raw(*args):
    """Wait for the raw value of a telmetry item against a condition or for a timeout
    and then check against the condition
    Supports two signatures:
    wait_check_raw(target_name, packet_name, item_name, comparison_to_eval, timeout, polling_rate)
    or
    wait_check_raw('target_name packet_name item_name > 1', timeout, polling_rate)"""
    return _wait_check(True, *args)


def _wait_check_tolerance(raw, *args):
    type_string = "wait_check_tolerance"
    if raw:
        type_string += "_raw"
    if raw:
        type = "RAW"
    else:
        type = "CONVERTED"
    (
        target_name,
        packet_name,
        item_name,
        expected_value,
        tolerance,
        timeout,
        polling_rate,
    ) = wait_tolerance_process_args(args, type_string)
    start_time = time.time()
    value = telemetry.tlm_variable(target_name, packet_name, item_name, type)
    if isinstance(value, list):
        expected_value, tolerance = array_tolerance_process_args(
            len(value), expected_value, tolerance, type_string
        )
        success, value = cosmos_script_wait_implementation_array_tolerance(
            len(value),
            target_name,
            packet_name,
            item_name,
            type,
            expected_value,
            tolerance,
            timeout,
            polling_rate,
        )
        time_float = time.time() - start_time

        message = ""
        for i in range(0, len(value)):
            range_bottom = expected_value[i] - tolerance[i]
            range_top = expected_value[i] + tolerance[i]
            check_str = "WAIT: {:s}[{:d}]".format(
                _upcase(target_name, packet_name, item_name), i
            )
            range_str = "range {:g} to {:g} with value == {:g} after waiting {:g} seconds".format(
                range_bottom, range_top, value[i], time_float
            )
            if value[i] >= range_bottom and value[i] <= range_top:
                message += "{:s} was within #{:s}\n".format(check_str, range_str)
            else:
                message += "{:s} failed to be within {:s}\n".format(
                    check_str, range_str
                )

        if success:
            LOGGER.info(message)
        else:
            raise CosmosCheckError(message)
    else:
        success, value = cosmos_script_wait_implementation_tolerance(
            target_name,
            packet_name,
            item_name,
            type,
            expected_value,
            tolerance,
            timeout,
            polling_rate,
        )
        time_float = time.time() - start_time
        range_bottom = expected_value - tolerance
        range_top = expected_value + tolerance
        check_str = "CHECK: {:s}".format(_upcase(target_name, packet_name, item_name))
        range_str = (
            "range {:g} to {:g} with value == {:g} after waiting {:g} seconds".format(
                range_bottom, range_top, value, time_float
            )
        )
        if success:
            LOGGER.info("{:s} was within {:s}".format(check_str, range_str))
        else:
            message = "{:s} failed to be within {:s}".format(check_str, range_str)
            raise CosmosCheckError(message)
    return time_float


def wait_check_tolerance(*args):
    _wait_check_tolerance(False, *args)


def wait_check_tolerance_raw(*args):
    _wait_check_tolerance(True, *args)


def wait_check_expression(
    exp_to_eval, timeout, polling_rate=DEFAULT_TLM_POLLING_RATE, context=None
):
    """Wait on an expression to be true.  On a timeout, the script will pause"""
    start_time = time.time()
    success = cosmos_script_wait_implementation_expression(
        exp_to_eval, timeout, polling_rate, context
    )
    time_float = time.time() - start_time
    if success:
        LOGGER.info(
            "CHECK: {:s} is TRUE after waiting {:g} seconds".format(
                exp_to_eval, time_float
            )
        )
    else:
        message = "CHECK: {:s} is FALSE after waiting {:g} seconds".format(
            exp_to_eval, time_float
        )
        raise CosmosCheckError(message)
    return time_float


def wait_expression_stop_on_timeout(*args):
    return wait_check_expression(*args)


def _wait_packet(
    check,
    target_name,
    packet_name,
    num_packets,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
):
    """Wait for a telemetry packet to be received a certain number of times or timeout"""
    if check:
        type = "CHECK"
    else:
        type = "WAIT"
    initial_count = telemetry.tlm(target_name, packet_name, "RECEIVED_COUNT")
    start_time = time.time()
    success, value = cosmos_script_wait_implementation(
        target_name,
        packet_name,
        "RECEIVED_COUNT",
        "CONVERTED",
        ">= {:d}".format(initial_count + num_packets),
        timeout,
        polling_rate,
    )
    time_float = time.time() - start_time
    if success:
        LOGGER.info(
            "{:s}: {:s} {:s} received {:d} times after waiting {:g} seconds".format(
                type,
                target_name.upper(),
                packet_name.upper(),
                value - initial_count,
                time_float,
            )
        )
    else:
        message = "{:s}: {:s} {:s} expected to be received {:d} times but only received {:d} times after waiting {:g} seconds".format(
            type,
            target_name.upper(),
            packet_name.upper(),
            num_packets,
            value - initial_count,
            time_float,
        )
        if check:
            raise CosmosCheckError(message)
        else:
            LOGGER.warning(message)
    return time_float


def wait_packet(
    target_name,
    packet_name,
    num_packets,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
):
    return _wait_packet(
        False, target_name, packet_name, num_packets, timeout, polling_rate
    )


def wait_check_packet(
    target_name,
    packet_name,
    num_packets,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
):
    """Wait for a telemetry packet to be received a certain number of times or timeout and raise an error"""
    return _wait_packet(
        True, target_name, packet_name, num_packets, timeout, polling_rate
    )


##########################################
# Protected Methods
##########################################


def check_process_args(args, function_name):
    length = len(args)
    if length == 1:
        (
            target_name,
            packet_name,
            item_name,
            comparison_to_eval,
        ) = extract.extract_fields_from_check_text(args[0])
    elif length == 4:
        target_name = args[0]
        packet_name = args[1]
        item_name = args[2]
        comparison_to_eval = args[3]
    else:
        # Invalid number of arguments
        raise RuntimeError(
            "ERROR: Invalid number of arguments ({:d}) passed to {:s}()".format(
                len(args), function_name
            )
        )
    return [target_name, packet_name, item_name, comparison_to_eval]


def check_tolerance_process_args(args, function_name):
    length = len(args)
    if length == 3:
        target_name, packet_name, item_name = extract.extract_fields_from_tlm_text(
            args[0]
        )
        expected_value = args[1]
        tolerance = abs(args[2])
    elif length == 5:
        target_name = args[0]
        packet_name = args[1]
        item_name = args[2]
        expected_value = args[3]
        tolerance = abs(args[4])
    else:
        # Invalid number of arguments
        raise RuntimeError(
            "ERROR: Invalid number of arguments ({:d}) passed to {:s}()".format(
                length, function_name
            )
        )
    return [target_name, packet_name, item_name, expected_value, tolerance]


def _execute_wait(
    target_name,
    packet_name,
    item_name,
    value_type,
    comparison_to_eval,
    timeout,
    polling_rate,
):
    start_time = time.time()
    success, value = cosmos_script_wait_implementation(
        target_name,
        packet_name,
        item_name,
        value_type,
        comparison_to_eval,
        timeout,
        polling_rate,
    )
    time_float = time.time() - start_time
    wait_str = "WAIT: {:s} {:s}".format(
        _upcase(target_name, packet_name, item_name), comparison_to_eval
    )
    value_str = "with value == {:s} after waiting {:g} seconds".format(
        str(value), time_float
    )
    if success:
        LOGGER.info("{:s} success {:s}".format(wait_str, value_str))
    else:
        LOGGER.warning("{:s} failed {:s}".format(wait_str, value_str))


def wait_process_args(args, function_name, value_type):
    time_float = None

    length = len(args)
    if length == 0:
        start_time = time.time()
        cosmos_script_sleep()
        time_float = time.time() - start_time
        LOGGER.info(
            "WAIT: Indefinite for actual time of {:g} seconds".format(time_float)
        )

    elif length == 1:
        try:
            value = float(args[0])
        except ValueError:
            raise RuntimeError("Non-numeric wait time specified")

        start_time = time.time()
        cosmos_script_sleep(value)
        time_float = time.time() - start_time
        LOGGER.info(
            "WAIT: {:g} seconds with actual time of {:g} seconds".format(
                value, time_float
            )
        )

    elif length == 2 or length == 3:
        (
            target_name,
            packet_name,
            item_name,
            comparison_to_eval,
        ) = extract.extract_fields_from_check_text(args[0])
        timeout = args[1]
        if length == 3:
            polling_rate = args[2]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
        _execute_wait(
            target_name,
            packet_name,
            item_name,
            value_type,
            comparison_to_eval,
            timeout,
            polling_rate,
        )

    elif length == 5 or length == 6:
        target_name = args[0]
        packet_name = args[1]
        item_name = args[2]
        comparison_to_eval = args[3]
        timeout = args[4]
        if length == 6:
            polling_rate = args[5]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
        _execute_wait(
            target_name,
            packet_name,
            item_name,
            value_type,
            comparison_to_eval,
            timeout,
            polling_rate,
        )
    else:
        # Invalid number of arguments
        raise RuntimeError(
            "ERROR: Invalid number of arguments ({:d}) passed to {:s}()".format(
                length, function_name
            )
        )
    return time_float


def wait_tolerance_process_args(args, function_name):
    length = len(args)
    if length == 4 or length == 5:
        target_name, packet_name, item_name = extract.extract_fields_from_tlm_text(
            args[0]
        )
        expected_value = args[1]
        tolerance = abs(args[2])
        timeout = args[3]
        if length == 5:
            polling_rate = args[4]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
    elif length == 6 or length == 7:
        target_name = args[0]
        packet_name = args[1]
        item_name = args[2]
        expected_value = args[3]
        tolerance = abs(args[4])
        timeout = args[5]
        if length == 7:
            polling_rate = args[6]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
    else:
        # Invalid number of arguments
        raise RuntimeError(
            "ERROR: Invalid number of arguments ({:d}) passed to {:s}()".format(
                length, function_name
            )
        )
    return [
        target_name,
        packet_name,
        item_name,
        expected_value,
        tolerance,
        timeout,
        polling_rate,
    ]


def array_tolerance_process_args(array_size, expected_value, tolerance, function_name):
    """
    When testing an array with a tolerance, the expected value and tolerance
    can both be supplied as either an array or a single value.  If a single
    value is passed in, that value will be used for all array elements.
    """
    if isinstance(expected_value, list):
        if array_size != len(expected_value):
            raise RuntimeError(
                "ERROR: Invalid array size for expected_value passed to {:s}()".format(
                    function_name
                )
            )
    else:
        expected_value = [expected_value] * array_size
    if isinstance(tolerance, list):
        if array_size != len(tolerance):
            raise RuntimeError(
                "ERROR: Invalid array size for tolerance passed to {:s}()".format(
                    function_name
                )
            )
    else:
        tolerance = [tolerance] * array_size
    return [expected_value, tolerance]


def wait_check_process_args(args, function_name):
    length = len(args)
    if length == 2 or length == 3:
        (
            target_name,
            packet_name,
            item_name,
            comparison_to_eval,
        ) = extract.extract_fields_from_check_text(args[0])
        timeout = args[1]
        if length == 3:
            polling_rate = args[2]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
    elif length == 5 or length == 6:
        target_name = args[0]
        packet_name = args[1]
        item_name = args[2]
        comparison_to_eval = args[3]
        timeout = args[4]
        if length == 6:
            polling_rate = args[5]
        else:
            polling_rate = DEFAULT_TLM_POLLING_RATE
    else:
        # Invalid number of arguments
        raise RuntimeError(
            "ERROR: Invalid number of arguments ({:d}) passed to {:s}()".format(
                length, function_name
            )
        )
    return [
        target_name,
        packet_name,
        item_name,
        comparison_to_eval,
        timeout,
        polling_rate,
    ]


def cosmos_script_sleep(sleep_time=None):
    """sleep in a script - returns true if canceled mid sleep"""
    if sleep_time != None:
        time.sleep(sleep_time)
    else:
        input("Infinite Wait - Press Enter to Continue: ")
    return False


def _cosmos_script_wait_implementation(
    target_name, packet_name, item_name, value_type, timeout, polling_rate, exp_to_eval
):
    end_time = time.time() + timeout

    while True:
        work_start = time.time()
        value = telemetry.tlm_variable(target_name, packet_name, item_name, value_type)
        if eval(exp_to_eval):
            return [True, value]
        if time.time() >= end_time:
            break

        delta = time.time() - work_start
        sleep_time = polling_rate - delta
        end_delta = end_time - time.time()
        if end_delta < sleep_time:
            sleep_time = end_delta
        if sleep_time < 0:
            sleep_time = 0
        canceled = cosmos_script_sleep(sleep_time)

        if canceled:
            value = telemetry.tlm_variable(
                target_name, packet_name, item_name, value_type
            )
            if eval(exp_to_eval):
                return [True, value]
            else:
                return [False, value]

    return [False, value]


# Wait for a converted telemetry item to pass a comparison
def cosmos_script_wait_implementation(
    target_name,
    packet_name,
    item_name,
    value_type,
    comparison_to_eval,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
):
    exp_to_eval = "value " + comparison_to_eval
    return _cosmos_script_wait_implementation(
        target_name,
        packet_name,
        item_name,
        value_type,
        timeout,
        polling_rate,
        exp_to_eval,
    )


def cosmos_script_wait_implementation_tolerance(
    target_name,
    packet_name,
    item_name,
    value_type,
    expected_value,
    tolerance,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
):
    exp_to_eval = "(value >= ({:g} - {:g}) and value <= ({:g} + {:g}))".format(
        expected_value, abs(tolerance), abs(tolerance), expected_value
    )
    return _cosmos_script_wait_implementation(
        target_name,
        packet_name,
        item_name,
        value_type,
        timeout,
        polling_rate,
        exp_to_eval,
    )


def cosmos_script_wait_implementation_array_tolerance(
    array_size,
    target_name,
    packet_name,
    item_name,
    value_type,
    expected_value,
    tolerance,
    timeout,
    polling_rate=DEFAULT_TLM_POLLING_RATE,
):
    statements = []
    for i in range(array_size):
        statements.append(
            "(value >= ({:g} - {:g}) and value <= ({:g} + {:g}))".format(
                expected_value[i],
                abs(tolerance[i]),
                abs(tolerance[i]),
                expected_value[i],
            )
        )
    exp_to_eval = " and ".join(statements)
    return _cosmos_script_wait_implementation(
        target_name,
        packet_name,
        item_name,
        value_type,
        timeout,
        polling_rate,
        exp_to_eval,
    )


def cosmos_script_wait_implementation_expression(
    exp_to_eval, timeout, polling_rate, locals=None
):
    """Wait on an expression to be true."""
    end_time = time.time() + timeout
    # ~ context = ScriptRunnerFrame.instance.script_binding if !context and defined? ScriptRunnerFrame and ScriptRunnerFrame.instance

    while True:
        work_start = time.time()
        if eval(exp_to_eval, locals):
            return True
        if time.time() >= end_time:
            break

        delta = time.time() - work_start
        sleep_time = polling_rate - delta
        end_delta = end_time - time.time()
        if end_delta < sleep_time:
            sleep_time = end_delta
        if sleep_time < 0:
            sleep_time = 0
        canceled = cosmos_script_sleep(sleep_time)

        if canceled:
            if eval(exp_to_eval, locals):
                return True
            else:
                return None

    return None


def check_eval(target_name, packet_name, item_name, comparison_to_eval, value):
    string = "value " + comparison_to_eval
    check_str = "CHECK: {:s} {:s}".format(
        _upcase(target_name, packet_name, item_name), comparison_to_eval
    )
    value_str = "with value == {:s}".format(str(value))
    if eval(string):
        LOGGER.info("{:s} success {:s}".format(check_str, value_str))
    else:
        message = "{:s} failed {:s}".format(check_str, value_str)
        raise CosmosCheckError(message)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
decorators.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import time
import functools
import requests
import logging

from cosmosc2.environment import MAX_RETRY_COUNT
from cosmosc2.exceptions import CosmosError, CosmosRetryError
from cosmosc2.__version__ import __title__

logger = logging.getLogger(__title__)


def request_wrapper(func):
    @functools.wraps(func)
    def _request(*args, **kwargs):
        try:
            value = func(*args, **kwargs)
            return value
        except ValueError as exc:
            err = f"ValueError {exc} while requesting token."
            raise CosmosError(err) from exc
        except requests.Timeout as exc:
            err = f"Timeout error while requesting {exc.request.url!r}"
            raise CosmosRetryError(err) from exc
        except requests.HTTPError as exc:
            err = f"Error response {exc.response.status_code} while requesting {exc.request.url!r}."
            if 400 >= exc.response.status_code < 500:
                raise CosmosError(err) from exc
            elif 500 >= exc.response.status_code < 600:
                raise CosmosRetryError(err) from exc
        except requests.RequestException as exc:
            err = f"An error occurred while requesting {exc.request.url!r}."
            raise CosmosError(err) from exc

    return _request


def retry_wrapper(func):
    @functools.wraps(func)
    def _request(*args, **kwargs):
        exception = None
        for i in range(0, MAX_RETRY_COUNT):
            try_ = f"request {i} out of {MAX_RETRY_COUNT}"
            try:
                logger.debug("%s", try_)
                value = func(*args, **kwargs)
                return value
            except CosmosRetryError as exc:
                logger.error("%s failed with a retryable error, %s", try_, exc)
                exception = exc
                time.sleep(i ** 4)
            except CosmosError as exc:
                logger.error("%s request failed with an error, %s", try_, exc)
                exception = exc
                break
        raise CosmosError("failed all attempts to pull data") from exception

    return _request



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
exceptions.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt


class CosmosError(RuntimeError):
    """ """


class CosmosConnectionError(CosmosError):
    """
    TODO
    """


class CosmosRetryError(CosmosError):
    """
    TODO
    """


class CosmosCheckError(CosmosError):
    """
    TODO
    """


class CosmosRequestError(CosmosError):
    """
    CosmosRequestError

    Parameters:
        message (str): The Request Error from Cosmos v5
        request (cosmosc2.execptions.CosmosJsonRpcRequest): CosmosJsonRpcRequest v5
    """

    def __init__(self, message: str, request):
        super().__init__(message, request)
        self.request = request


class CosmosResponseError(CosmosError):
    """
    CosmosResponseError

    Parameters:
        request (cosmosc2.execptions.CosmosJsonRpcRequest): CosmosJsonRpcRequest v5
        response (cosmosc2.execptions.CosmosJsonRpcErrorResponse): CosmosJsonRpcErrorResponse v5
    """

    def __init__(self, request, response):
        super().__init__(request, response)
        self.request = request
        self.response = response



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
__version__.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

__title__ = "cosmosc2"
__description__ = "Python Support for Ball Aerospace COSMOS v5"
__url__ = "https://github.com/BallAerospace/python-cosmosc2"
__version__ = "1.2.0"



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
limits.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import cosmosc2


def get_out_of_limits():
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("get_out_of_limits")


def get_overall_limits_state(ignored_items=None):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("get_overall_limits_state", ignored_items)


def limits_enabled(*args):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("limits_enabled?", *args)


def enable_limits(*args):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("enable_limits", *args)


def disable_limits(*args):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("disable_limits", *args)


def get_stale(with_limits_only=False, target_name=None):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("get_stale", with_limits_only, target_name)


def get_limits(target_name, packet_name, item_name, limits_set=None):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request(
        "get_limits", target_name, packet_name, item_name, limits_set
    )


def set_limits(
    target_name,
    packet_name,
    item_name,
    red_low,
    yellow_low,
    yellow_high,
    red_high,
    green_low=None,
    green_high=None,
    limits_set="CUSTOM",
    persistence=None,
    enabled=True,
):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request(
        "set_limits",
        target_name,
        packet_name,
        item_name,
        red_low,
        yellow_low,
        yellow_high,
        red_high,
        green_low,
        green_high,
        limits_set,
        persistence,
        enabled,
    )


def get_limits_groups():
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("get_limits_groups")


def enable_limits_group(group_name):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("enable_limits_group", group_name)


def disable_limits_group(group_name):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("disable_limits_group", group_name)


def get_limits_sets():
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("get_limits_sets")


def get_current_limits_set():
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("get_current_limits_set")


def set_limits_set(limits_set):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("set_limits_set", limits_set)


def get_limits_set():
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("get_limits_set")


def subscribe_limits_events(queue_size=1000):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("subscribe_limits_events", queue_size)


def unsubscribe_limits_events(id_):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("unsubscribe_limits_events", id_)


def get_limits_event(id_, non_block=False):
    """
    TODO
    """
    return cosmosc2.COSMOS.json_rpc_request("get_limits_event", id_, non_block)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
json_rpc/__init__.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

from cosmosc2.environment import JSON_RPC_VERSION


class CosmosJsonRpc(dict):
    """Base class for all JSON Remote Procedure Calls.

    Provides basic comparison and Hash to JSON conversions.
    """

    def __init__(self):
        super().__init__()
        self["jsonrpc"] = JSON_RPC_VERSION

    @property
    def id(self):
        return self.get("id")

    @property
    def json_rpc(self):
        return self.get("jsonrpc", JSON_RPC_VERSION)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
connection.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

from contextlib import ContextDecorator
import json
import logging
from requests import Session
from requests.auth import AuthBase
from threading import RLock, Event, Thread
import requests

from cosmosc2.__version__ import __title__
from cosmosc2.authorization import generate_auth
from cosmosc2.environment import *
from cosmosc2.exceptions import CosmosConnectionError
from cosmosc2.decorators import request_wrapper, retry_wrapper
from cosmosc2.json_rpc import CosmosJsonRpcRequest, CosmosJsonRpcResponse

LOGGER = logging.getLogger(__title__)


class CosmosConnection(ContextDecorator):
    """Class to perform JSON-RPC Calls to the COSMOS Server (or other JsonDrb server)

    The CosmosConnection can be used to call COSMOS server methods directly:
      server = CosmosConnection(hostname: "127.0.0.1", port: 7777)
      server.write(...)
    """

    def __init__(
        self,
        schema: str = COSMOS_API_SCHEMA,
        hostname: str = COSMOS_API_HOSTNAME,
        port: int = COSMOS_API_PORT,
        timeout: float = 5.0,
        scope: str = COSMOS_SCOPE,
        auth: AuthBase = None,
    ) -> None:
        """Constructor

        Parameters:
        hostname -- The name of the machine which has started the JSON service
        port -- The port number of the JSON service
        timeout -- The amount of time the socket will read until an error
        scope -- The scope or project the connection will add to the request
        """
        self.id = 0
        self.scope = scope
        self.timeout = float(timeout)
        self.request_url = f"{schema}://{hostname}:{port}"
        self._session = Session()
        self._session.headers = {
            "User-Agent": USER_AGENT,
        }
        self.auth = generate_auth() if auth is None else auth
        self._mutex = RLock()
        self._shutdown_needed = Event()

    def shutdown(self):
        """Permanently disconnects from the JSON server"""
        self._shutdown_needed.set()
        self._session.close()

    def update_kwargs(self, request_kwargs: dict):
        request_kwargs["auth"] = self.auth
        params = request_kwargs.get("params", {})
        params["scope"] = self.scope
        request_kwargs["params"] = params

    def json_rpc_request(self, method_name, *args, **kwargs):
        """Forwards all method calls to the remote service.

        method_name -- Name of the method to call
        args -- Array of parameters to pass to the method
        kwargs -- Dict of parameters to pass to the method
        return -- The result of the method call. If something goes wrong with the
          protocol a exception extended from RuntimeError is raised.
        """
        if self._shutdown_needed.is_set():
            raise CosmosConnectionError("shutdown event is set, exiting")
        with self._mutex:
            self.id += 1
            json_rpc_request = CosmosJsonRpcRequest(
                self.id, method_name, self.scope, *args, **kwargs
            )
            resp = self._make_json_rpc_request(json_rpc_request.to_hash())
            try:
                json_rpc_response = CosmosJsonRpcResponse.from_bytes(resp.content)
                LOGGER.debug(
                    "response %s %s", type(json_rpc_response), json_rpc_response
                )
                return json_rpc_response.result
            except AttributeError:
                return json_rpc_response

    @request_wrapper
    @retry_wrapper
    def _make_json_rpc_request(self, hash_: dict):
        """Use the python requests libary to send the request to Cosmos.

        This is an internal method that uses two decorators. request_wrapper
        captures errors from the request libary. retry_wrapper captures exceptions
        from the requst_wrapper and will retry or bubble up the error.
        return -- bytes: request.content
            https://docs.python-requests.org/en/master/user/quickstart/#binary-response-content
        """
        request_kwargs = {
            "auth": self.auth,
            "url": f"{self.request_url}/cosmos-api/api",
            "data": json.dumps(hash_),
            "headers": {
                "Content-Type": "application/json-rpc",
            },
        }
        LOGGER.debug("calling with %s", request_kwargs)
        resp = self._session.post(**request_kwargs)
        LOGGER.debug(
            "resp: %s total_seconds: %f content: %s",
            resp,
            resp.elapsed.total_seconds(),
            resp.content,
        )
        return resp

    @request_wrapper
    @retry_wrapper
    def get(self, endpoint: str, **kwargs):
        """Use the python requests libary to send the request to Cosmos.

        This is an internal method that uses two decorators. request_wrapper
        captures errors from the request libary. retry_wrapper captures exceptions
        from the requst_wrapper and will retry or bubble up the error.
        return -- bytes: request.content
            https://docs.python-requests.org/en/master/user/quickstart/#binary-response-content
        """
        kwargs["url"] = f"{self.request_url}{endpoint}"
        self.update_kwargs(kwargs)
        LOGGER.debug("calling with %s", kwargs)
        resp = requests.get(**kwargs)
        LOGGER.debug(
            "resp: %s total_seconds: %f content: %s",
            resp,
            resp.elapsed.total_seconds(),
            resp.content,
        )
        return resp

    @request_wrapper
    @retry_wrapper
    def post(self, endpoint: str, **kwargs):
        """Use the python requests libary to send the request to Cosmos.

        This is an internal method that uses two decorators. request_wrapper
        captures errors from the request libary. retry_wrapper captures exceptions
        from the requst_wrapper and will retry or bubble up the error.
        return -- bytes: request.content
            https://docs.python-requests.org/en/master/user/quickstart/#binary-response-content
        """
        kwargs["url"] = f"{self.request_url}{endpoint}"
        self.update_kwargs(kwargs)
        LOGGER.debug("calling with %s", kwargs)
        resp = requests.post(**kwargs)
        LOGGER.debug(
            "resp: %s total_seconds: %f content: %s",
            resp,
            resp.elapsed.total_seconds(),
            resp.content,
        )
        return resp

    @request_wrapper
    @retry_wrapper
    def put(self, endpoint: str, **kwargs):
        """Use the python requests libary to send the request to Cosmos.

        This is an internal method that uses two decorators. request_wrapper
        captures errors from the request libary. retry_wrapper captures exceptions
        from the requst_wrapper and will retry or bubble up the error.
        return -- bytes: request.content
            https://docs.python-requests.org/en/master/user/quickstart/#binary-response-content
        """
        kwargs["url"] = f"{self.request_url}{endpoint}"
        self.update_kwargs(kwargs)
        LOGGER.debug("calling with %s", kwargs)
        resp = requests.put(**kwargs)
        LOGGER.debug(
            "resp: %s total_seconds: %f content: %s",
            resp,
            resp.elapsed.total_seconds(),
            resp.content,
        )
        return resp

    @request_wrapper
    @retry_wrapper
    def delete(self, endpoint: str, **kwargs):
        """Use the python requests libary to send the request to Cosmos.

        This is an internal method that uses two decorators. request_wrapper
        captures errors from the request libary. retry_wrapper captures exceptions
        from the requst_wrapper and will retry or bubble up the error.
        return -- bytes: request.content
            https://docs.python-requests.org/en/master/user/quickstart/#binary-response-content
        """
        kwargs["url"] = f"{self.request_url}{endpoint}"
        self.update_kwargs(kwargs)
        LOGGER.debug("calling with %s", kwargs)
        resp = requests.delete(**kwargs)
        LOGGER.debug(
            "resp: %s total_seconds: %f content: %s",
            resp,
            resp.elapsed.total_seconds(),
            resp.content,
        )
        return resp



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
cmd_tlm_server.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import cosmosc2

DEFAULT_SERVER_MESSAGES_QUEUE_SIZE = 1000


def get_interface_names():
    """The get_interface_names method returns a list of the interfaces in the system in an array.
    Syntax / Example:
        interface_names = get_interface_names()
    """
    return cosmosc2.COSMOS.json_rpc_request("get_interface_names")


def connect_interface(interface_name, *params):
    """The connect_interface method connects to targets associated with a COSMOS interface.
    Syntax:
        connect_interface("<Interface Name>", <Interface Parameters (optional)>)
    """
    return cosmosc2.COSMOS.json_rpc_request(
        "connect_interface", interface_name, *params
    )


def disconnect_interface(interface_name):
    """The disconnect_interface method disconnects from targets associated with a COSMOS interface.
    Syntax:
        disconnect_interface("<Interface Name>")
    """
    return cosmosc2.COSMOS.json_rpc_request("disconnect_interface", interface_name)


def get_router_names():
    """The get_router_names method returns a list of the routers in the
    system in an array.
    Syntax:
        router_names = get_router_names()
    """
    return cosmosc2.COSMOS.json_rpc_request("get_router_names")


def get_all_router_info():
    """The get_all_router_info method returns information about all routers.
    The return value is an array of arrays where each subarray contains the
    router name, connection state, number of connected clients, transmit queue
    size, receive queue size, bytes transmitted, bytes received, packets
    received, and packets sent.
    Syntax:
        router_info = get_all_router_info()
    """
    return cosmosc2.COSMOS.json_rpc_request("get_all_router_info")


def connect_router(router_name, *params):
    """The connect_router method connects a COSMOS router.
    Syntax:
        connect_router("<Router Name>", <Router Parameters (optional)>)
    """
    return cosmosc2.COSMOS.json_rpc_request("connect_router", router_name, *params)


def disconnect_router(router_name):
    """The disconnect_router method disconnects a COSMOS router.
    Syntax:
        disconnect_router("<Router Name>")
    """
    return cosmosc2.COSMOS.json_rpc_request("disconnect_router", router_name)


def get_all_target_info():
    """The get_all_target_info method returns information about all targets.
    The return value is an array of arrays where each subarray contains the
    target name, interface name, command count, and telemetry count for a target.
    Syntax:
        target_info = get_all_target_info()
    """
    return cosmosc2.COSMOS.json_rpc_request("get_all_target_info")


def get_target_ignored_parameters(target_name):
    """ """
    return cosmosc2.COSMOS.json_rpc_request(
        "get_target_ignored_parameters", target_name
    )


def get_target_ignored_items(target_name):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_target_ignored_items", target_name)


def get_interface_info(interface_name):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_interface_info", interface_name)


def get_all_interface_info():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_all_interface_info")


def get_router_info(router_name):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_router_info", router_name)


def get_all_cmd_info():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_all_cmd_info")


def get_all_tlm_info():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_all_tlm_info")


def get_cmd_cnt(target_name, command_name):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_cmd_cnt", target_name, command_name)


def get_tlm_cnt(target_name, packet_name):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_tlm_cnt", target_name, packet_name)


def get_packet_loggers():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_packet_loggers")


def get_packet_logger_info(packet_logger_name):
    """ """
    return cosmosc2.COSMOS.json_rpc_request(
        "get_packet_logger_info", packet_logger_name
    )


def get_all_packet_logger_info():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_all_packet_logger_info")


def get_background_tasks():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_background_tasks")


def start_background_task(task_name):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("start_background_task", task_name)


def stop_background_task(task_name):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("stop_background_task", task_name)


def get_server_status():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_server_status")


def get_cmd_log_filename(packet_log_writer_name="DEFAULT"):
    """ """
    return cosmosc2.COSMOS.json_rpc_request(
        "get_cmd_log_filename", packet_log_writer_name
    )


def get_tlm_log_filename(packet_log_writer_name="DEFAULT"):
    """ """
    return cosmosc2.COSMOS.json_rpc_request(
        "get_tlm_log_filename", packet_log_writer_name
    )


def start_logging(packet_log_writer_name="ALL", label=None):
    """ """
    return cosmosc2.COSMOS.json_rpc_request(
        "start_logging", packet_log_writer_name, label
    )


def stop_logging(packet_log_writer_name="ALL"):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("stop_logging", packet_log_writer_name)


def start_cmd_log(packet_log_writer_name="ALL", label=None):
    """ """
    return cosmosc2.COSMOS.json_rpc_request(
        "start_cmd_log", packet_log_writer_name, label
    )


def start_tlm_log(packet_log_writer_name="ALL", label=None):
    """ """
    return cosmosc2.COSMOS.json_rpc_request(
        "start_tlm_log", packet_log_writer_name, label
    )


def stop_cmd_log(packet_log_writer_name="ALL"):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("stop_cmd_log", packet_log_writer_name)


def stop_tlm_log(packet_log_writer_name="ALL"):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("stop_tlm_log", packet_log_writer_name)


def start_raw_logging_interface(interface_name="ALL"):
    """ """
    return cosmosc2.COSMOS.json_rpc_request(
        "start_raw_logging_interface", interface_name
    )


def stop_raw_logging_interface(interface_name="ALL"):
    """ """
    return cosmosc2.COSMOS.json_rpc_request(
        "stop_raw_logging_interface", interface_name
    )


def start_raw_logging_router(router_name="ALL"):
    """The start_raw_logging_router method starts logging of raw data on one or all routers. This is for debugging purposes only.
    Syntax:
        start_raw_logging_router("<Router Name (optional)>")
    """
    return cosmosc2.COSMOS.json_rpc_request("start_raw_logging_router", router_name)


def stop_raw_logging_router(router_name="ALL"):
    """The stop_raw_logging_router method stops logging of raw data on one or all routers. This is for debugging purposes only.
    Syntax:
        stop_raw_logging_router("<Router Name (optional)>")
    """
    return cosmosc2.COSMOS.json_rpc_request("stop_raw_logging_router", router_name)


def get_server_message_log_filename():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_server_message_log_filename")


def start_new_server_message_log():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("start_new_server_message_log")


def subscribe_server_messages(queue_size=DEFAULT_SERVER_MESSAGES_QUEUE_SIZE):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("subscribe_server_messages", queue_size)


def unsubscribe_server_messages(id_):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("unsubscribe_server_messages", id_)


def get_server_message(id_, non_block=False):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_server_message", id_, non_block)


def cmd_tlm_reload():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("cmd_tlm_reload")


def cmd_tlm_clear_counters():
    """ """
    return cosmosc2.COSMOS.json_rpc_request("cmd_tlm_clear_counters")


def get_output_logs_filenames(filter_="*tlm.bin"):
    """ """
    return cosmosc2.COSMOS.json_rpc_request("get_output_logs_filenames", filter_)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
stream_api.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import json

from cosmosc2.authorization import CosmosAuthorization, generate_auth
from cosmosc2.environment import COSMOS_SCOPE
from cosmosc2.stream import CosmosAsyncStream


class CosmosAsyncClient:
    def __init__(
        self,
        stream: CosmosAsyncStream,
        scope: str = COSMOS_SCOPE,
        auth: CosmosAuthorization = None,
    ) -> None:
        self.stream = stream
        self.scope = scope
        self.auth = generate_auth() if auth is None else auth

    def streaming_id(self):
        """
        generate the streaming id to send on the websocket.

        Returns:
            str: json string for identifier.
        """
        return json.dumps(
            {
                "channel": "StreamingChannel",
                "scope": self.scope,
                "token": self.auth.get(),
            }
        )

    def streaming_channel_sub(self, callback):
        """
        Have the stream subscribe to the streaming channel

        Parameters:
            callback (callable): methods called when data recivied
        """
        self.stream.subscribe(
            "/cosmos-api/cable",
            {"command": "subscribe", "identifier": self.streaming_id()},
            callback,
        )

    def streaming_channel_unsub(self):
        """
        Have the stream unsubscribe from the streaming channel
        """
        self.stream.queue(
            "/cosmos-api/cable",
            {"command": "unsubscribe", "identifier": self.streaming_id()},
        )
        self.stream.unsubscribe("/cosmos-api/cable")

    def streaming_channel_add(
        self,
        items: list,
        mode: str = "DECOM",
        start_time: int = None,
        end_time: int = None,
    ):
        """
        Have the stream already subscribed to the streaming channel
        add items with a time range to the stream.

        Parameters:
            items (list):
            mode (str):
            start_time (int):
            end_time (int):
        """
        data = json.dumps(
            {
                "scope": self.scope,
                "token": self.auth.get(),
                "mode": mode,
                "items": items,
                "start_time": start_time,
                "end_time": end_time,
                "action": "add",
            }
        )
        self.stream.queue(
            "/cosmos-api/cable",
            {
                "command": "message",
                "identifier": self.streaming_id(),
                "data": data,
            },
        )

    def message_id(self, history_count: int):
        """
        generate the message id to send on the websocket.

        Returns:
            history_count (str): stream history to request
        """
        return json.dumps(
            {
                "channel": "MessagesChannel",
                "scope": self.scope,
                "token": self.auth.get(),
                "history_count": history_count,
            }
        )

    def message_channel_sub(self, history_count, callback):
        """
        Have the stream unsubscribe from the streaming channel

        Parameters:
            history_count (int): stream history to request
            callback (callable): methods called when data recivied
        """
        self.stream.subscribe(
            "/cosmos-api/cable",
            {"command": "subscribe", "identifier": self.message_id(history_count)},
            callback,
        )

    def message_channel_unsub(self, history_count: int):
        """
        Have the stream unsubscribe from the streaming channel

        Parameters:
            history_count (int): stream history to request
        """
        self.stream.queue(
            "/cosmos-api/cable",
            {"command": "unsubscribe", "identifier": self.message_id(history_count)},
        )
        self.stream.unsubscribe("/cosmos-api/cable")



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
stream.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import asyncio
import json
import logging
from threading import Thread
import websockets

from cosmosc2.environment import *

logger = logging.getLogger("websockets")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


class CosmosAsyncError(RuntimeError):
    pass


class CosmosAsyncStop(StopAsyncIteration):
    pass


class CosmosAsyncStream(Thread):
    def __init__(
        self,
        schema: str = COSMOS_WS_SCHEMA,
        hostname: str = COSMOS_API_HOSTNAME,
        port: int = COSMOS_API_PORT,
    ):
        """
        Is the base thread class to access cosmos websockets.

        Parameters:
            schema (str): [optional] will default to environment variable
            hostname (str): [optional] will default to environment variable
            port (int): [optional] will default to environment variable
        """
        super().__init__()
        self._tasks = {}
        self._events = {}
        self._queues = {}
        self._loop = asyncio.new_event_loop()
        self._stop_event = asyncio.Event(loop=self._loop)
        self._url = f"{schema}://{hostname}:{port}"

    def run(self):
        """
        Should be called from the thread.start() method.
        """
        try:
            self._loop.run_until_complete(self._stop_event.wait())
            self._loop.run_until_complete(self._clean())
        finally:
            self._loop.close()

    def stop(self):
        """
        Should be called to stop all websockets.
        """
        self._loop.call_soon_threadsafe(self._stop_event.set)

    async def _clean(self):
        """
        Should wait for websockets to shutdown
        """
        for task in self._tasks.values():
            await asyncio.wait_for(task, timeout=5)
        await asyncio.gather(*self._tasks.values(), loop=self._loop)

    def subscribe(self, endpoint, sub_msg, callback):
        """
        Start a new asyncio.Task to the cosmos websocket endpoints.

        Parameters:
            endpoint (str): example: /cosmos-api/cable
            sub_msg (dict): send to cosmos once connected
            callback (callable): method to return messages to
        """

        def _subscribe():
            if endpoint not in self._tasks:
                listen = self._listen(endpoint, sub_msg, callback)
                self._tasks[endpoint] = self._loop.create_task(listen)
                self._queues[endpoint] = asyncio.Queue()
                self._events[endpoint] = asyncio.Event(loop=self._loop)

        self._loop.call_soon_threadsafe(_subscribe)

    def unsubscribe(self, endpoint):
        """
        Stop an asyncio.Task

        Parameters:
            endpoint (str): example: /cosmos-api/cable
        """

        def _unsubscribe():
            event = self._events.pop(endpoint, None)
            if event is not None:
                event.set()

        self._loop.call_soon_threadsafe(_unsubscribe)

    def queue(self, endpoint, message):
        """
        Queue a message to send to the websocket.

        Parameters:
            endpoint (str): example: /cosmos-api/cable
            message (dict): json based message to send
        """

        def _queue():
            queue = self._queues.get(endpoint, None)
            if queue is not None:
                queue.put_nowait(message)

        self._loop.call_soon_threadsafe(_queue)

    async def _listen(self, endpoint, sub_msg, callback):
        """
        Base asyncio.Task to connect and manage websocket.

        Parameters:
            endpoint (str): example: /cosmos-api/cable
            sub_msg (dict): json based message to send
            callback (callable): method/function to call with data
        """
        url = f"{self._url}{endpoint}"
        try:
            ws = await websockets.connect(url, loop=self._loop)
            await self._welcome(ws)
            await self._confirm(ws, sub_msg)
            await self._handle(endpoint, ws, callback)
            await ws.close()
        except asyncio.CancelledError:
            logging.info(f"{endpoint} has been canceled")
        except CosmosAsyncStop:
            logging.info(f"stopping {endpoint}")
        except CosmosAsyncError as e:
            logging.error(f"failed connection {endpoint}, {e}")
        except Exception as e:
            logging.exception(e)
        finally:
            logging.debug(f"exitting task: {endpoint}")
            self._queues.pop(endpoint, None)
            self._events.pop(endpoint, None)

    @staticmethod
    async def _welcome(ws):
        """
        Confirm welcome from ActionCable websocket.

        Parameters:
            ws (WebSocket): open websocket
        """
        data = await ws.recv()
        data = json.loads(data)
        if data["type"] != "welcome":
            raise CosmosAsyncError("failed to get welcome message")

    @staticmethod
    async def _confirm(ws, sub_msg):
        """
        Send sub_msg to ActionCable websocket.

        Parameters:
            ws (WebSocket): open websocket
            sub_msg (dict): json based object
        """
        json_msg = json.dumps(sub_msg)
        logging.debug(f"sending: {json_msg}")
        await ws.send(json_msg)
        data = await ws.recv()
        data = json.loads(data)
        logging.debug(f"recv: {data}")

    async def _handle(self, endpoint, ws, callback):
        """
        Once connect send messages queued to ActionCable websocket
        and receive messages from.

        Parameters:
            endpoint (str): endpoint of connection
            ws (WebSocket): open websocket
            callback (callable): function/method to send data to
        """
        queue = self._queues[endpoint]
        event = self._events[endpoint]
        while event.is_set() is False and self._stop_event.is_set() is False:
            await self._send(queue, ws)
            data = await ws.recv()
            data = json.loads(data)
            callback(data)

    @staticmethod
    async def _send(queue, ws):
        """
        Send a message in the queue.

        Parameters:
            queue (Queue): Outgoing message queue
            ws (WebSocket): open websocket
        """
        try:
            message = queue.get_nowait()
            if message is None:
                raise CosmosAsyncStop()
            else:
                logging.debug(f"sending: {message}")
                await ws.send(json.dumps(message))
        except asyncio.QueueEmpty:
            pass



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
timeline_api.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import cosmosc2


def cosmos_timelines():
    """Get the cosmos timeline api.
    Syntax / Example:
        timelines = cosmos_timelines()
    """
    resp = cosmosc2.COSMOS.get(
        "/cosmos-api/timeline", headers={"Accept": "application/json"}
    )
    return resp.json()


def cosmos_timeline_activities(timeline: str):
    """Get the cosmos health api.
    Syntax / Example:
        activities = cosmos_timeline_activities("alpha")
    """
    resp = cosmosc2.COSMOS.get(
        f"/cosmos-api/timeline/{timeline}/activities",
        headers={"Accept": "application/json"},
    )
    return resp.json()


def cosmos_timeline_activity_count(timeline: str):
    """Get the cosmos timeline activity count.
    Syntax / Example:
        count = cosmos_timeline_activity_count("alpha")
    """
    resp = cosmosc2.COSMOS.get(
        f"/cosmos-api/timeline/{timeline}/count", headers={"Accept": "plain/txt"}
    )
    return resp.text



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
extract.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import re


SCANNING_REGULAR_EXPRESSION = re.compile(
    r"(?:\"(?:[^\\\"]|\\.)*\") | (?:'(?:[^\\']|\\.)*') | (?:\[.*\]) | \S+", re.VERBOSE
)

SPLIT_WITH_REGEX = re.compile(r"\s+with\s+", re.IGNORECASE)

# Regular expression to identify a String as a floating point number
FLOAT_CHECK_REGEX = re.compile(r"\A\s*[-+]?\d*\.\d+\s*\Z")

# Regular expression to identify a String as a floating point number in
# scientific notation
SCIENTIFIC_CHECK_REGEX = re.compile(
    r"\A\s*[-+]?(\d+((\.\d+)?)|(\.\d+))[eE][-+]?\d+\s*\Z"
)

# Regular expression to identify a String as an integer
INT_CHECK_REGEX = re.compile(r"\A\s*[-+]?\d+\s*\Z")

# Regular expression to identify a String as an integer in hexadecimal format
HEX_CHECK_REGEX = re.compile(r"\A\s*0[xX][\dabcdefABCDEF]+\s*\Z")

# Regular expression to identify a String as an Array of numbers
ARRAY_CHECK_REGEX = re.compile(r"\A\s*\[.*\]\s*\Z")


def remove_quotes(string: str):
    """Returns the string with leading and trailing quotes removed"""
    if (string.startswith('"') and string.endswith('"')) or (
        string.startswith("'") and string.endswith("'")
    ):
        return string[1:-1]
    return string


def is_float(string):
    """Returns whether the String represents a floating point number"""
    if FLOAT_CHECK_REGEX.match(string) or SCIENTIFIC_CHECK_REGEX.match(string):
        return True
    return False


def is_int(string):
    """Returns whether the String represents an integer"""
    if INT_CHECK_REGEX.match(string):
        return True
    return False


def is_hex(string):
    """Whether the String represents a hexadecimal number"""
    if HEX_CHECK_REGEX.match(string):
        return True
    return False


def is_array(string):
    """Whether the String represents an Array"""
    if ARRAY_CHECK_REGEX.match(string):
        return True
    return False


def convert_to_value(string):
    """Converts the String into either a Float, Integer, or Array
    depending on what the String represents. It can successfully convert
    floating point numbers in both fixed and scientific notation, integers
    in hexadecimal notation, and Arrays. If it can not be converted into
    any of the above then the original String is returned.
    """
    return_value = string
    try:
        if is_float(string):
            # Floating Point in normal or scientific notation
            return_value = float(string)
        elif is_int(string):
            # Integer
            return_value = int(string)
        elif is_hex(string):
            # Hex
            return_value = int(string)
        elif isinstance(string, list):
            # Array
            return_value = eval(string)
    except Exception:
        # Something went wrong so just return the string as is
        pass
    return return_value


def hex_to_byte_string(string):
    """Converts the String representing a hexadecimal number (i.e. "0xABCD")
    to a binary String with the same data (i.e "\xAB\xCD")
    @return [String] Binary byte string"""

    # Remove leading 0x or 0X
    if string[:2] == "0x" or string[:2] == "0X":
        string = string[2:]

    return bytearray.fromhex(string)


def add_cmd_parameter(keyword, value, cmd_params):
    quotes_removed = remove_quotes(value)
    if value == quotes_removed:
        cmd_params[keyword] = convert_to_value(value)
    else:
        cmd_params[keyword] = quotes_removed


def extract_fields_from_cmd_text(text):
    split_string = re.split(SPLIT_WITH_REGEX, text, 2)
    if len(split_string) == 1 and SPLIT_WITH_REGEX.match(text):
        raise RuntimeError(
            "ERROR: 'with' must be followed by parameters : {:s}".format(text)
        )

    # Extract target_name and cmd_name
    first_half = split_string[0].split(" ")
    if len(first_half) < 2:
        raise RuntimeError(
            "ERROR: Both Target Name and Command Name must be given : {:s}".format(text)
        )
    if len(first_half) > 2:
        raise RuntimeError(
            "ERROR: Only Target Name and Command Name must be given before 'with' : {:s}".format(
                text
            )
        )
    target_name = first_half[0]
    cmd_name = first_half[1]
    cmd_params = {}

    if len(split_string) == 2:
        # Extract Command Parameters
        second_half = SCANNING_REGULAR_EXPRESSION.findall(split_string[1])
        keyword = None
        value = None
        comma = None
        for item in second_half:
            if not keyword:
                keyword = item
                continue
            if not value:
                if item.endswith(","):
                    value = item[0:-1]
                    comma = True
                else:
                    value = item
                    continue
            if not comma:
                if item != ",":
                    raise RuntimeError(
                        "Missing comma in command parameters: {:s}".format(text)
                    )
            add_cmd_parameter(keyword, value, cmd_params)
            keyword = None
            value = None
            comma = None
        if keyword:
            if value:
                add_cmd_parameter(keyword, value, cmd_params)
            else:
                raise RuntimeError(
                    "Missing value for last command parameter: {:s}".format(text)
                )

    return [target_name, cmd_name, cmd_params]


def extract_fields_from_tlm_text(text):
    split_string = text.split(" ")
    if len(split_string) != 3:
        raise RuntimeError(
            "ERROR: Telemetry Item must be specified as 'TargetName PacketName ItemName' : {:s}".format(
                text
            )
        )
    target_name = split_string[0]
    packet_name = split_string[1]
    item_name = split_string[2]
    return [target_name, packet_name, item_name]


def extract_fields_from_set_tlm_text(text):
    error_msg = "ERROR: Set Telemetry Item must be specified as 'TargetName PacketName ItemName = Value' : {:s}".format(
        text
    )
    # We have to handle these cases:
    # set_tlm("TGT PKT ITEM='new item'")
    # set_tlm("TGT PKT ITEM = 'new item'")
    # set_tlm("TGT PKT ITEM= 'new item'")
    # set_tlm("TGT PKT ITEM ='new item'")
    split_string = text.split("=")
    if len(split_string) < 2 or not split_string[1].strip():
        raise error_msg
    split_string = (
        split_string[0].strip().split(" ") + "=".join(split_string[1:]).strip()
    )
    if len(split_string) != 4:  # Ensure tgt,pkt,item,value
        raise error_msg
    target_name = split_string[0]
    packet_name = split_string[1]
    item_name = split_string[2]
    value = convert_to_value(split_string[3].strip())
    if isinstance(value, str):
        value = remove_quotes(value)
    return [target_name, packet_name, item_name, value]


def extract_fields_from_check_text(text):
    split_string = text.split(" ")
    if len(split_string) < 3:
        raise RuntimeError("ERROR: Check improperly specified: {:s}")
    target_name = split_string[0]
    packet_name = split_string[1]
    item_name = split_string[2]
    comparison_to_eval = None
    if len(split_string) == 3:
        return [target_name, packet_name, item_name, comparison_to_eval]
    if len(split_string) < 4:
        raise RuntimeError("ERROR: Check improperly specified: {:s}".format(text))
    comparison_to_eval = " ".join(split_string[3:])
    if split_string[3] == "=":
        raise RuntimeError("ERROR: Use '==' instead of '=': {:s}")
    return [target_name, packet_name, item_name, comparison_to_eval]



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
# -*- coding: latin-1 -*-
"""
json_rpc/response.py
"""

# Copyright 2022 Ball Aerospace & Technologies Corp.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt

import json

from cosmosc2.environment import JSON_RPC_VERSION
from cosmosc2.exceptions import CosmosError
from cosmosc2.json_rpc.base import CosmosJsonRpc
from cosmosc2.json_rpc.error import CosmosJsonRpcError


class CosmosJsonRpcResponse(CosmosJsonRpc):
    """Represents a JSON Remote Procedure Call Response"""

    def __init__(self, id_):
        """Constructor

        Parameters:
        id_ -- The identifier which will be matched to the request
        """
        super().__init__()
        self["id"] = id_

    @classmethod
    def from_bytes(cls, response_data: bytes):
        """Creates a JsonRpcResponse object from a JSON encoded String.

        The version must be 2.0 and the JSON must include the id members. It
        must also include either result for success or error for failure but
        never both.

        Parameters:
        response_data -- JSON encoded string representing the response
        """

        msg = "invalid json-rpc {} response".format(JSON_RPC_VERSION)
        try:
            hash_ = json.loads(response_data.decode("latin-1"))
        except Exception as e:
            raise CosmosError(msg, response_data) from e

            # Verify the jsonrpc version is correct and there is an ID
        if hash_.get("jsonrpc") != JSON_RPC_VERSION:
            raise CosmosError(msg, response_data)

        try:
            return CosmosJsonRpcErrorResponse.from_hash(hash_)
        except KeyError:
            pass

        try:
            return CosmosJsonRpcSuccessResponse.from_hash(hash_)
        except KeyError:
            pass

        raise CosmosError(msg, response_data)


def convert_json_class(object_):
    if isinstance(object_, dict):
        try:
            json_class = object_["json_class"]
            raw = object_["raw"]
            if json_class == "Float":
                if raw == "Infinity":
                    return float("inf")
                elif raw == "-Infinity":
                    return -float("inf")
                elif raw == "NaN":
                    return float("nan")
                return raw
            return bytearray(raw)
        except Exception:
            for key, value in object_.items():
                object_[key] = convert_json_class(value)
            return object_
    elif isinstance(object_, (tuple, list)):
        object_ = list(object_)
        index = 0
        for value in object_:
            object_[index] = convert_json_class(value)
            index += 1
        return object_
    else:
        return object_


class CosmosJsonRpcSuccessResponse(CosmosJsonRpcResponse):
    """Represents a JSON Remote Procedure Call Success Response"""

    def __init__(self, id_, result):
        """Constructor

        Parameters:
        id_ -- The identifier which will be matched to the request
        result -- The result of the request
        """
        super().__init__(id_)
        result = convert_json_class(result)
        self["result"] = result

    @property
    def result(self):
        """Return the result of the method request"""
        return self["result"]

    @classmethod
    def from_hash(cls, hash_):
        """Creates a JsonRpcSuccessResponse object from a Hash

        Parameters
        hash_ -- Hash containing the following keys: result and id
        """
        return cls(hash_["id"], hash_["result"])


class CosmosJsonRpcErrorResponse(CosmosJsonRpcResponse):
    """Represents a JSON Remote Procedure Call Error Response"""

    def __init__(self, id_, error):
        """Constructor

        Parameters:
        id -- The identifier which will be matched to the request
        error -- The error object
        """
        super().__init__(id_)
        self["error"] = CosmosJsonRpcError.from_hash(error)

    @property
    def error(self):
        """Returns the error object"""
        return self["error"]

    @classmethod
    def from_hash(cls, hash_):
        """Creates a JsonRpcErrorResponse object from a Hash

        Parameters:
        hash -- Hash containing the following keys: error and id
        """
        return cls(hash_["id"], hash_["error"])



import os
import sys

# See cosmosc2/docs/environment.md for environment documentation

os.environ["COSMOS_API_USER"] = "operator"
os.environ["COSMOS_API_PASSWORD"] = "operator"
os.environ["COSMOS_LOG_LEVEL"] = "INFO"
os.environ["COSMOS_API_SCHEMA"] = "http"
os.environ["COSMOS_API_HOSTNAME"] = "127.0.0.1"
os.environ["COSMOS_API_PORT"] = "2900"

from cosmosc2 import *

print(cosmos_status())
print(cosmos_health())

# ~ # telemetry.py
print(tlm("INST HEALTH_STATUS TEMP1"))
print(tlm_raw("INST HEALTH_STATUS TEMP1"))
print(tlm_formatted("INST HEALTH_STATUS TEMP1"))
print(tlm_with_units("INST HEALTH_STATUS TEMP1"))
print(tlm_variable("INST HEALTH_STATUS TEMP1", "RAW"))
print(set_tlm("INST HEALTH_STATUS TEMP1 = 5"))
print(get_tlm_packet("INST", "HEALTH_STATUS"))
print(
    get_tlm_values(
        [
            "INST__HEALTH_STATUS__TEMP1__CONVERTED",
            "INST__HEALTH_STATUS__TEMP2__CONVERTED",
        ]
    )
)
print(get_target_list())
print(get_target("INST"))
print(get_tlm_buffer("INST", "HEALTH_STATUS"))

id_ = subscribe_packets([["INST", "HEALTH_STATUS"]])
print(id_)

sys.exit(1)

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
# commands.py
cmd("INST ABORT")
cmd_no_range_check("INST COLLECT with TYPE NORMAL, TEMP 50.0")
cmd_no_hazardous_check("INST CLEAR")
cmd_no_checks("INST COLLECT with TYPE SPECIAL, TEMP 50.0")
cmd_raw("INST COLLECT with TYPE 0, TEMP 10.0")
cmd_raw_no_range_check("INST COLLECT with TYPE 0, TEMP 50.0")
cmd_raw_no_hazardous_check("INST CLEAR")
cmd_raw_no_checks("INST COLLECT with TYPE 1, TEMP 50.0")
send_raw("EXAMPLE_INT", "\x00\x00\x00\x00")
send_raw_file("EXAMPLE_INT", os.path.join(FILE_PATH, "test.txt"))
get_cmd_list("INST")
get_cmd_param_list("INST", "COLLECT")
get_cmd_hazardous("INST", "CLEAR")
get_cmd_value("INST", "COLLECT", "TEMP")
get_cmd_time()
get_cmd_buffer("INST", "COLLECT")
cmd_no_range_check("INST COLLECT with TYPE NORMAL, TEMP 50.0")


# timeline_api.py
print(cosmos_timelines())

update_scope("UPDATE")

shutdown()
